"""Choose largest number out of several represented in various formats."""

from dataclasses import dataclass
from random import Random
from typing import Any, Optional

from ..coaching import BaseCurriculum, RangeAttributeDefinition, ScalarAttributeDefinition
from ..factory import ProceduralDataset, register_dataset

QUESTION_TEMPLATE = """Your task is to pick the largest/smallest number out of several options.

Your output should be only the number of interest.

Now, pick the {size} number of the following candidates: {numbers}
"""

DATASET_NAME = "number_format"


@dataclass
class NumberFormatConfig:
    """Configuration for Count Bits dataset generation"""

    min_num_candidates: int = 2  # Minimum number of candidates
    max_num_candidates: int = 5  # Maximum number of candidates
    min_n: float = 1_000  # Lower bound for the numbers
    max_n: float = 1_000_000_000  # Upper bound for the numbers
    max_delta: float = 10.0

    size: int = 500  # Virtual dataset size
    seed: Optional[int] = None

    def validate(self):
        """Validate configuration parameters"""
        assert 2 <= self.min_num_candidates, "min_num_candidates must be at least 2"
        assert (
            self.min_num_candidates <= self.max_num_candidates
        ), "min_num_candidates must be less than max_num_candidates"
        assert 1 <= self.min_n, "min_n must be at least 1"
        assert self.min_n < self.max_n, "min_n must be less than max_n"
        assert 0 < self.max_delta, "max_delta must be greater than 0"


class NumberFormatDataset(ProceduralDataset):
    """Generates Count Bits exercises with configurable difficulty"""

    def __init__(self, config: NumberFormatConfig):
        super().__init__(config=config, seed=config.seed, size=config.size)

    def _get_candidates(self, rng: Random, num_candidates: int) -> list:
        """Generate a list of candidates"""
        base = round(rng.uniform(self.config.min_n, self.config.max_n), 3)
        candidates = [base]
        for _ in range(num_candidates - 1):
            delta = round(rng.uniform(-self.config.max_delta, self.config.max_delta), 3)
            candidates.append(base + delta)
        return candidates

    def _transform_candidates(self, rng: Random, candidates: list[float]) -> list[str]:
        """Randomly apply different number formats to the candidates"""
        output = []
        for candidate in candidates:
            format_type = rng.choice(["standard", "english", "scientific"])
            if format_type == "standard":
                output.append(f"{candidate:f}")
            elif format_type == "english":
                output.append(f"{candidate:,}")
            elif format_type == "scientific":
                output.append(f"{candidate:.15e}")
        return output

    def score_answer(self, answer: Optional[str], entry: dict[str, Any]) -> float:
        """Overwrite this method in derived classes if a single oracle answer is not available."""
        oracle_answer = entry["metadata"]["solution"]
        if isinstance(answer, str) and len(answer) > 0:
            try:
                answer = float(answer.strip().replace(",", ""))
                if abs(answer - oracle_answer) < 1e-2:
                    return 1.0
            except:
                pass
        return 0.0

    def __getitem__(self, idx: int) -> dict:
        """Generate a single Count Bits question"""
        rng = Random(self.seed + idx)

        num_candidates = rng.randint(self.config.min_num_candidates, self.config.max_num_candidates)
        candidates = self._get_candidates(rng, num_candidates)
        formatted_candidates = self._transform_candidates(rng, candidates)

        size = rng.choice(["largest", "smallest"])
        answer = max(candidates) if size == "largest" else min(candidates)

        return {
            "question": QUESTION_TEMPLATE.format(numbers=" ".join(formatted_candidates), size=size),
            "answer": str(answer),
            "metadata": {
                "source_dataset": DATASET_NAME,
                "source_index": idx,
                "candidates": candidates,
                "solution": answer,
                "formatted_candidates": formatted_candidates,
                "size": size,
                "num_candidates": num_candidates,
                "difficulty": {
                    "num_candidates": (self.config.min_num_candidates, self.config.max_num_candidates),
                    "n": (self.config.min_n, self.config.max_n),
                    "min_delta": self.config.max_delta,
                },
            },
        }


class NumberFormatCurriculum(BaseCurriculum):
    def __init__(self):
        super().__init__(NumberFormatCurriculum.__name__, NumberFormatConfig)

        self._define_attributes(
            RangeAttributeDefinition(
                name="num_candidates",
                levels=[5, 25, 100, 500],
                description="Number of candidates",
                lower_field_name="min_num_candidates",
                upper_field_name="max_num_candidates",
                ensure_interval=True,
            ),
            RangeAttributeDefinition(
                name="n",
                levels=[1_000, 100_000, 1_000_000, 1_000_000_000],
                description="Magnitude of the values",
                lower_field_name="min_n",
                upper_field_name="max_n",
                ensure_interval=True,
            ),
            ScalarAttributeDefinition(
                name="max_delta",
                field_name="max_delta",
                levels=[1e1, 1e0, 1e-3, 1e-6],
                description="Max delta",
            ),
        )


register_dataset(DATASET_NAME, NumberFormatDataset, NumberFormatConfig, NumberFormatCurriculum)



import random
from random import Random
from typing import Any

NUM_OF_PAIRS_GENERATED = 5


def create_random_list(rng: Random):
    length = rng.randint(3, 10)
    return [rng.randint(1, 100) for _ in range(length)]


def create_list_of_fives(rng: Random):
    length = rng.randint(1, 7)  # Random length between 1 and 7
    return [5] * length


def sort_integers(lst, order="ascending"):
    """
    Sorts a list of integers in ascending or descending order.

    Parameters:
        lst (list): The list of integers to sort.
        order (str): The order to sort in. Options are 'ascending' or 'descending'.

    Returns:
        list: The sorted list.
    """
    if order == "ascending":
        return sorted(lst)  # Sort in ascending order
    elif order == "descending":
        return sorted(lst, reverse=True)  # Sort in descending order
    else:
        raise ValueError("Invalid order. Use 'ascending' or 'descending'.")


def create_random_odd_numbers(count, start, end):
    """
    Generates a list of random odd numbers.

    Parameters:
        count (int): The number of odd numbers to generate.
        start (int): The lower bound of the range (inclusive).
        end (int): The upper bound of the range (inclusive).

    Returns:
        list: A list of random odd numbers.
    """
    odd_numbers = []
    while len(odd_numbers) < count:
        num = random.randint(start, end)  # Generate a random number
        if num % 2 != 0:  # Check if the number is odd
            odd_numbers.append(num)
    return odd_numbers


def create_numbers_divisible_by_five_or_ten(rng: Random):
    result = []
    for i in range(NUM_OF_PAIRS_GENERATED):
        if i % 2 == 0:
            num = create_random_odd_numbers(1, 1, 1000)[0] * 5  # Random multiple of 5
        else:
            num = rng.randint(1, 100) * 10  # Random multiple of 10
        result.append(num)
    return result


def generate_0(rng: Random) -> dict[str, Any]:
    """Generate input and output pairs where input remains unchanged"""
    pairs = {}

    for _ in range(NUM_OF_PAIRS_GENERATED):
        input = create_random_list(rng)
        input = str(input)
        output = input
        pairs[input] = output

    return pairs


def generate_1(rng: Random) -> dict[str, Any]:
    """Generate input and output pairs where output is a list of the third element
    after removing all other elements
    """
    pairs = {}

    for _ in range(NUM_OF_PAIRS_GENERATED):
        input = create_random_list(rng)
        target_idx = 2
        output = [input[target_idx]]
        input = str(input)
        output = str(output)
        pairs[input] = output

    return pairs


def generate_2(rng: Random) -> dict[str, Any]:
    """Generate input and output pairs where output is a reversed list of the input"""
    pairs = {}
    for _ in range(NUM_OF_PAIRS_GENERATED):
        input = create_random_list(rng)
        output = input[::-1]
        input = str(input)
        output = str(output)
        pairs[input] = output

    return pairs


def generate_3(rng: Random) -> dict[str, Any]:
    """Generate input and output pairs where output is the sum of unique elements in the list less than 30"""
    pairs = {}
    for _ in range(NUM_OF_PAIRS_GENERATED):
        input = create_random_list(rng)
        unique_input = list(set(input))

        total_sum = 0
        for num in unique_input:
            if num < 30:
                total_sum += num

        input = str(input)
        output = str([total_sum])
        pairs[input] = output

    return pairs


def generate_4(rng: Random) -> dict[str, Any]:
    """Generate input and output pairs where output is the count of elements equal to 5"""
    pairs = {}
    for i in range(NUM_OF_PAIRS_GENERATED):
        input = create_random_list(rng)

        if i % 2 == 0:
            input += create_list_of_fives(rng)

        # Shuffle the new input with fives
        rng.shuffle(input)

        total_count = 0
        for num in input:
            if num == 5:
                total_count += 1

        input = str(input)
        output = str([total_count])
        pairs[input] = output

    return pairs


def generate_5(rng: Random) -> dict[str, Any]:
    """Generate input and output pairs where output is a list of elements that are followed by an even number

    NOTE: This is suppose to be a relatively hard problem
    """
    pairs = {}
    for i in range(NUM_OF_PAIRS_GENERATED):
        input = create_random_list(rng)
        output = []
        for i in range(1, len(input)):

            # If the current element is an even number, we then add previous element into output
            if input[i] % 2 == 0:
                output.append(input[i - 1])

        input = str(input)
        output = str(output)
        pairs[input] = output

    return pairs


def generate_6(rng: Random) -> dict[str, Any]:
    """Generate input and output pairs where output is a list of elements where each element in input is added to its position(Using zero-indexing)"""
    pairs = {}
    for i in range(NUM_OF_PAIRS_GENERATED):
        input = create_random_list(rng)
        output = []
        for i, num in enumerate(input):
            element = i + num
            output.append(element)

        input = str(input)
        output = str(output)
        pairs[input] = output

    return pairs


def generate_7(rng: Random) -> dict[str, Any]:
    """Generate input and output pairs where output is a list of element whose position is indicated by the last element in the input

    EXAMPLE:
    1. [26, 88, 60, 1, 17, 75, 97, 89, 1] -> [88]
    2.  [49, 71, 2, 61, 3]: [61]
    """
    pairs = {}
    for _ in range(NUM_OF_PAIRS_GENERATED):
        input = create_random_list(rng)
        # Create a chosen index between the bounds of the size of the input
        chosen_index = rng.randint(0, len(input) - 1)
        # Replace the last element with chosen_index
        input[-1] = chosen_index
        output = [input[chosen_index]]

        input = str(input)
        output = str(output)
        pairs[input] = output

    return pairs


def generate_8(rng: Random) -> dict[str, Any]:
    """Generate input and output pairs where output is count of elements in the input"""
    pairs = {}
    for _ in range(NUM_OF_PAIRS_GENERATED):
        input = create_random_list(rng)
        output = [len(input)]

        input = str(input)
        output = str(output)
        pairs[input] = output

    return pairs


def generate_9(rng: Random) -> dict[str, Any]:
    """Generate input and output pairs where output is sum total of elements in the input"""
    pairs = {}
    for _ in range(NUM_OF_PAIRS_GENERATED):
        input = create_random_list(rng)
        output = [sum(input)]

        input = str(input)
        output = str(output)
        pairs[input] = output

    return pairs


def generate_10(rng: Random) -> dict[str, Any]:
    """Generate input and output pairs where output is a list of the elements in ascending order"""
    pairs = {}
    for _ in range(NUM_OF_PAIRS_GENERATED):
        input = create_random_list(rng)
        output = sort_integers(input, order="ascending")

        input = str(input)
        output = str(output)
        pairs[input] = output

    return pairs


def generate_11(rng: Random) -> dict[str, Any]:
    """Generate input and output pairs where output is a list of the elements in descending order"""
    pairs = {}
    for _ in range(NUM_OF_PAIRS_GENERATED):
        input = create_random_list(rng)
        output = sort_integers(input, order="descending")

        input = str(input)
        output = str(output)
        pairs[input] = output

    return pairs


def generate_12(rng: Random) -> dict[str, Any]:
    """Generate input and output pairs where output is a list of the elements where the first and last element in input are replaced by their
    successor. Example, for an integer 4, successor is 5
    """
    pairs = {}
    for _ in range(NUM_OF_PAIRS_GENERATED):
        input = create_random_list(rng)
        # Create successor for first and last element using a copy of input
        output = input.copy()
        output[0] += 1
        output[-1] += 1

        input = str(input)
        output = str(output)
        pairs[input] = output

    return pairs


def generate_13(rng: Random) -> dict[str, Any]:
    """Generate input and output pairs where output is [1] if list of input elements is in ascending order, [0] in descending order"""
    pairs = {}
    for i in range(NUM_OF_PAIRS_GENERATED):
        input = create_random_list(rng)
        if i % 2 == 0:
            input = sort_integers(input, order="ascending")
            output = [1]
        else:
            input = sort_integers(input, order="descending")
            output = [0]

        input = str(input)
        output = str(output)
        pairs[input] = output

    return pairs


def generate_14(rng: Random) -> dict[str, Any]:
    """Generate input and output pairs where output is [1] if input element is divisible by 10, [0] if divisible by 5"""
    pairs = {}

    nums = create_numbers_divisible_by_five_or_ten(rng)
    for num in nums:
        if num % 10 == 0:
            input = [num]
            output = [1]
        else:
            input = [num]
            output = [0]

        input = str(input)
        output = str(output)
        pairs[input] = output

    return pairs


def generate_15(rng: Random) -> dict[str, Any]:
    """Generate input and output pairs where output is a twice the amount of last element in the input"""
    pairs = {}
    for _ in range(NUM_OF_PAIRS_GENERATED):
        starter_input = create_random_list(rng)
        length = len(starter_input)
        first_element = rng.choice(starter_input)
        input = [first_element]

        for _ in range(1, length):
            prev = input[-1]
            input.append(prev * 2)

        # Create output here to prevent building on strings
        output = str([input[-1] * 2])
        input = str(input)
        pairs[input] = output

    return pairs


def generate_16(rng: Random) -> dict[str, Any]:
    """Generate input and output pairs where output is built from a function 2x - 4
    NOTE: This is suppose to be amazingly hard for the LLM.
    """
    pairs = {}
    for _ in range(NUM_OF_PAIRS_GENERATED):
        starter_input = create_random_list(rng)
        first_element = rng.choice(starter_input)
        output = (2 * first_element) - 4
        input = str([first_element])
        pairs[input] = str([output])

    return pairs



from .advanced_geometry import AdvancedGeometryConfig, AdvancedGeometryCurriculum, AdvancedGeometryDataset
from .simple_geometry import SimpleGeometryConfig, SimpleGeometryCurriculum, SimpleGeometryDataset

__all__ = [
    "SimpleGeometryConfig",
    "SimpleGeometryDataset",
    "SimpleGeometryCurriculum",
    "AdvancedGeometryConfig",
    "AdvancedGeometryDataset",
    "AdvancedGeometryCurriculum",
]



"""
Pathfinding problems in a path-star graph structure.
Inspired by https://arxiv.org/pdf/2403.06963
"""

import random
from dataclasses import dataclass
from typing import Any, Optional

from ..coaching import BaseCurriculum, RangeAttributeDefinition, ScalarAttributeDefinition
from ..factory import ProceduralDataset, register_dataset

DATASET_NAME = "path_star"

PROMPT_TEMPLATE = """
Find a path from the start node to the goal node in the following path-star graph.
Respond with only the sequence of node labels in the path, including the start and goal nodes.
Separate node labels with a single space.

The graph is represented as a list of edges, where each edge is defined by two node labels.
The edges are separated by a vertical bar '|'. Then, the start and goal nodes are specified after a slash '/'.

Example:
|1 2|1 3|2 4|3 5/1 5 = 1 3 5

Solve the following task:
{task}
"""


@dataclass
class PathStarConfig:
    """Configuration for Path Star dataset generation"""

    degree: int = 3
    node_range: int = 100_000
    min_path_length: int = 3
    max_path_length: int = 5

    reversed: bool = False

    size: int = 500  # Virtual dataset size
    seed: Optional[int] = None

    def validate(self) -> None:
        """Validate configuration parameters"""
        assert self.degree >= 2, "degree must be at least 2"
        assert self.min_path_length >= 2, "min_path_length must be at least 2"
        assert self.min_path_length <= self.max_path_length, "min_path_length must be <= max_path_length"
        assert (
            self.node_range > self.degree * self.max_path_length + 1
        ), "node_range must exceed degree * max_path_length + 1 for unique labels"


class PathStarDataset(ProceduralDataset):
    """Procedurally generates path-star graph problems."""

    def __init__(self, config: PathStarConfig):
        super().__init__(config=config, seed=config.seed, size=config.size)

    def score_answer(self, answer: Optional[str], entry: dict[str, Any]) -> float:
        """Score an answer. Path is unique in a star graph, so only exact match counts."""
        if not isinstance(answer, str) or len(answer.strip()) == 0:
            return 0.0
        # Normalize: strip, collapse whitespace
        answer_normalized = " ".join(answer.strip().split())
        oracle_normalized = " ".join(entry["answer"].strip().split())
        if answer_normalized == oracle_normalized:
            return 1.0
        return 0.0

    def __getitem__(self, idx: int) -> dict[str, Any]:
        rng = random.Random(self.seed + idx)

        cfg: PathStarConfig = self.config
        center = rng.randrange(cfg.node_range)
        path_length = rng.randint(cfg.min_path_length, cfg.max_path_length)

        # allocate unique node labels
        paths = []
        used = {center}
        for _ in range(cfg.degree):
            path = []
            for _ in range(path_length):
                n = rng.randrange(cfg.node_range)
                while n in used:
                    n = rng.randrange(cfg.node_range)
                used.add(n)
                path.append(n)
            paths.append(path)

        goal_path = rng.choice(paths)
        goal = goal_path[-1]

        # build edge list
        edges = [(center, p[0]) for p in paths]
        for p in paths:
            edges.extend(zip(p[:-1], p[1:]))
        rng.shuffle(edges)

        edges_str = "".join(f"|{u} {v}" for u, v in edges)
        if cfg.reversed:
            prefix = f"{edges_str}/{goal} {center} = "
        else:
            prefix = f"{edges_str}/{center} {goal} = "
        question = PROMPT_TEMPLATE.format(task=prefix)

        # gold path
        gold = [center] + goal_path
        if cfg.reversed:
            gold = list(reversed(gold))
        answer = " ".join(map(str, gold))

        return {
            "question": question,
            "answer": answer,
            "metadata": {
                "source_dataset": DATASET_NAME,
                "source_index": idx,
                "center": center,
                "goal": goal,
                "path_length": path_length,
                "goal_path": gold if not cfg.reversed else list(reversed(gold)),
                "difficulty": {
                    "degree": cfg.degree,
                    "node_range": cfg.node_range,
                    "path_length": (cfg.min_path_length, cfg.max_path_length),
                },
            },
        }


class PathStarCurriculum(BaseCurriculum):
    def __init__(self):
        super().__init__(PathStarCurriculum.__name__, PathStarConfig)

        # Define attributes
        self._define_attributes(
            ScalarAttributeDefinition(
                name="degree",
                levels=[2, 3, 4, 5],
                description="Degree of the graph",
                field_name="degree",
            ),
            ScalarAttributeDefinition(
                name="node_range",
                levels=[10_000, 50_000, 100_000, 200_000],
                description="Range of node labels",
                field_name="node_range",
            ),
            RangeAttributeDefinition(
                name="path_length",
                levels=[3, 5, 6, 7],
                description="Length of paths in the graph",
                lower_field_name="min_path_length",
                upper_field_name="max_path_length",
            ),
        )


register_dataset(DATASET_NAME, PathStarDataset, PathStarConfig, PathStarCurriculum)



"""Rotate a square matrix clockwise.

A popular Leetcode problem:
https://leetcode.com/problems/rotate-image/description/
"""

from copy import deepcopy
from dataclasses import dataclass
from random import Random
from typing import Optional

from ..coaching import BaseCurriculum, RangeAttributeDefinition
from ..factory import ProceduralDataset, register_dataset

QUESTION_TEMPLATE = """Given a square matrix, your job is to rotate it clockwise.

Your output should be a matrix in the same format as the input.

Rotate the matrix below by {degrees} degrees clockwise:
{matrix}
"""

DATASET_NAME = "rotate_matrix"


@dataclass
class RotateMatrixConfig:
    """Configuration for Rotate Matrix dataset generation"""

    min_n: int = 2  # Minimum size of the matrix
    max_n: int = 10  # Maximum size of the matrix
    min_rotations: int = 0  # Minimum number of rotations
    max_rotations: int = 10  # Maximum number of rotations (90 degrees each)

    size: int = 500  # Virtual dataset size
    seed: Optional[int] = None

    def validate(self):
        """Validate configuration parameters"""
        assert 2 <= self.min_n <= self.max_n, "min_n and max_n must be between 2 and 10"
        assert 0 <= self.min_rotations <= self.max_rotations, "min_rotations must be between 0 and max_rotations"


class RotateMatrixDataset(ProceduralDataset):
    """Generates Rotate Matrix exercises with configurable difficulty"""

    def __init__(self, config: RotateMatrixConfig):
        super().__init__(config=config, seed=config.seed, size=config.size)

    def _get_matrix(self, rng: Random, n: int) -> list[list[int]]:
        """Generate a random matrix"""
        numbers = list(rng.randint(0, 9) for _ in range(n**2))
        matrix = [numbers[i * n : (i + 1) * n] for i in range(n)]
        return matrix

    def _rot90(self, matrix: list[list[int]]) -> list[list[int]]:
        """quarter clockwise rotation"""
        return [list(row) for row in zip(*matrix[::-1])]

    def _get_rotated(self, matrix: list[list[int]], num_rotations: int) -> list[list[int]]:
        """Rotate the matrix K times by 90 degrees clockwise"""
        num_rotations %= 4
        output = deepcopy(matrix)
        for _ in range(num_rotations):
            output = self._rot90(output)
        return output

    def _matrix_to_str(self, matrix: list[list[int]]) -> str:
        """Get a string representation of the matrix"""
        return "\n".join(" ".join(str(x) for x in row) for row in matrix)

    def __getitem__(self, idx: int) -> dict:
        """Generate a single Rotate Matrix question"""
        rng = Random(self.seed + idx)

        n = rng.randint(self.config.min_n, self.config.max_n)
        matrix = self._get_matrix(rng, n)
        num_rotations = rng.randint(self.config.min_rotations, self.config.max_rotations)
        matrix_str = self._matrix_to_str(matrix)

        answer = self._get_rotated(matrix, num_rotations)
        answer_str = self._matrix_to_str(answer)

        return {
            "question": QUESTION_TEMPLATE.format(matrix=matrix_str, degrees=num_rotations * 90),
            "answer": answer_str,
            "metadata": {
                "source_dataset": DATASET_NAME,
                "source_index": idx,
                "matrix": matrix,
                "num_rotations": num_rotations,
                "solution": answer,
                "n": n,
                "difficulty": {
                    "n": (self.config.min_n, self.config.max_n),
                    "num_rotations": (self.config.min_rotations, self.config.max_rotations),
                },
            },
        }


class RotateMatrixCurriculum(BaseCurriculum):
    def __init__(self):
        super().__init__(RotateMatrixCurriculum.__name__, RotateMatrixConfig)

        # Define attributes
        self._define_attributes(
            RangeAttributeDefinition(
                name="n",
                levels=[10, 25, 50, 100],
                description="Size of the square matrix",
                lower_field_name="min_n",
                upper_field_name="max_n",
            ),
            RangeAttributeDefinition(
                name="num_rotations",
                levels=[1, 5, 10, 15, 20],
                description="Number of 90-degree rotations",
                lower_field_name="min_rotations",
                upper_field_name="max_rotations",
                ensure_interval=True,
            ),
        )


register_dataset(DATASET_NAME, RotateMatrixDataset, RotateMatrixConfig, RotateMatrixCurriculum)



"""
Arithmetic tasks for training reasoning capabilities:
"""

from .basic_arithmetic import BasicArithmeticCurriculum, BasicArithmeticDataset, BasicArithmeticDatasetConfig
from .bitwise_arithmetic import BitwiseArithmeticConfig, BitwiseArithmeticCurriculum, BitwiseArithmeticDataset
from .calendar_arithmetic import CalendarArithmeticConfig, CalendarArithmeticCurriculum, CalendarArithmeticDataset
from .chain_sum import ChainSumConfig, ChainSumDataset
from .count_bits import CountBitsConfig, CountBitsCurriculum, CountBitsDataset
from .decimal_arithmetic import DecimalArithmeticConfig, DecimalArithmeticCurriculum, DecimalArithmeticDataset
from .decimal_chain_sum import DecimalChainSumConfig, DecimalChainSumCurriculum, DecimalChainSumDataset
from .dice import DiceConfig, DiceCurriculum, DiceDataset
from .fraction_simplification import (
    FractionSimplificationConfig,
    FractionSimplificationCurriculum,
    FractionSimplificationDataset,
)
from .gcd import GCDConfig, GCDCurriculum, GCDDataset
from .gsm_symbolic.gsm_symbolic import GSMSymbolicDataset, GSMSymbolicDatasetConfig
from .lcm import LCMConfig, LCMCurriculum, LCMDataset
from .leg_counting import LegCountingConfig, LegCountingCurriculum, LegCountingDataset
from .number_format import NumberFormatConfig, NumberFormatCurriculum, NumberFormatDataset
from .power_function import PowerFunctionConfig, PowerFunctionCurriculum, PowerFunctionDataset
from .prime_factorization import PrimeFactorizationConfig, PrimeFactorizationCurriculum, PrimeFactorizationDataset
from .products import ProductsConfig, ProductsDataset
from .time_intervals import TimeIntervalsConfig, TimeIntervalsCurriculum, TimeIntervalsDataset

__all__ = [
    "BasicArithmeticDataset",
    "BasicArithmeticDatasetConfig",
    "BasicArithmeticCurriculum",
    "ChainSumDataset",
    "ChainSumConfig",
    "CalendarArithmeticConfig",
    "CalendarArithmeticDataset",
    "CalendarArithmeticCurriculum",
    "FractionSimplificationConfig",
    "FractionSimplificationDataset",
    "FractionSimplificationCurriculum",
    "GCDConfig",
    "GCDDataset",
    "GCDCurriculum",
    "LCMConfig",
    "LCMDataset",
    "LCMCurriculum",
    "LegCountingConfig",
    "LegCountingDataset",
    "LegCountingCurriculum",
    "PowerFunctionConfig",
    "PowerFunctionDataset",
    "PowerFunctionCurriculum",
    "PrimeFactorizationConfig",
    "PrimeFactorizationDataset",
    "PrimeFactorizationCurriculum",
    "ProductsDataset",
    "ProductsConfig",
    "GSMSymbolicDatasetConfig",
    "GSMSymbolicDataset",
    "TimeIntervalsConfig",
    "TimeIntervalsDataset",
    "TimeIntervalsCurriculum",
    "CountBitsConfig",
    "CountBitsDataset",
    "CountBitsCurriculum",
    "DiceConfig",
    "DiceDataset",
    "DiceCurriculum",
    "NumberFormatConfig",
    "NumberFormatDataset",
    "NumberFormatCurriculum",
    "DecimalArithmeticConfig",
    "DecimalArithmeticDataset",
    "DecimalArithmeticCurriculum",
    "DecimalChainSumCurriculum",
    "DecimalChainSumConfig",
    "DecimalChainSumDataset",
    "BitwiseArithmeticConfig",
    "BitwiseArithmeticDataset",
    "BitwiseArithmeticCurriculum",
]



#!/usr/bin/env python3
from .Exceptions import BFSemanticError, BFSyntaxError
from .FunctionCompiler import FunctionCompiler
from .Functions import check_function_exists, get_function_object, insert_function_object
from .General import get_literal_token_code, is_token_literal, unpack_literal_tokens_to_array_dimensions
from .Globals import (
    create_variable_from_definition,
    get_global_variables_size,
    get_variable_dimensions,
    get_variable_size,
    insert_global_variable,
)
from .Lexical_analyzer import analyze
from .LibraryFunctionCompiler import insert_library_functions
from .Optimizer import optimize
from .Parser import Parser
from .Token import Token

"""
This file is responsible for creating FunctionCompiler objects and global variables objects
And finally, return the code of the main function
"""


class Compiler:
    def __init__(self, code, optimize_code=False):
        tokens = analyze(code)
        if optimize_code:
            optimize(tokens)
        self.parser = Parser(tokens)

    # global variables and functions
    def create_function_object(self):
        # function: (INT | VOID) ID LPAREN expression_list RPAREN LBRACE statements RBRACE
        # returns function named tuple

        if self.parser.current_token().type not in [Token.VOID, Token.INT]:
            raise BFSemanticError(
                "Function return type can be either void or int, not '%s'" % str(self.parser.current_token())
            )

        self.parser.check_next_tokens_are([Token.ID, Token.LPAREN])

        # save all tokens of this function
        function_name = self.parser.next_token(next_amount=1).data
        RPAREN_index = self.parser.find_matching(
            starting_index=self.parser.current_token_index + 2
        )  # first find RPAREN
        self.parser.check_next_token_is(Token.LBRACE, starting_index=RPAREN_index)
        RBRACE_index = self.parser.find_matching(starting_index=RPAREN_index + 1)  # then find RBRACE

        # take all tokens between INT and RBRACE and pass them to function object
        function_tokens = self.parser.tokens[self.parser.current_token_index : RBRACE_index + 1]
        # skip function definition
        self.parser.advance_to_token_at_index(RBRACE_index + 1)

        function = FunctionCompiler(function_name, function_tokens)
        return function

    def compile_global_variable_definition(self):
        # INT ID (ASSIGN NUM | (LBRACK NUM RBRACK)+ (ASSIGN LBRACE ... RBRACE)?)? SEMICOLON
        # returns code that initializes this variable, and advances pointer according to variable size

        self.parser.check_current_tokens_are([Token.INT, Token.ID])
        ID_token = self.parser.next_token()
        variable = create_variable_from_definition(self.parser, advance_tokens=True)
        insert_global_variable(variable)

        # if this is set to False, then the compiler assumes that initially all cells are set to zero
        # if this is set to True, then the compiler zeros each cell before using it (may generate a lot of unnecessary BF code)
        ZERO_CELLS_BEFORE_USE = False

        code = "[-]" if ZERO_CELLS_BEFORE_USE else ""
        if get_variable_size(variable) > 1:  # its an array
            if self.parser.current_token().type == Token.SEMICOLON:
                # array definition - INT ID (LBRACK NUM RBRACK)+ SEMICOLON
                self.parser.advance_token()  # skip SEMICOLON
                code = (code + ">") * get_variable_size(variable)  # advance to after this variable
                return code
            elif self.parser.current_token().type == Token.ASSIGN and self.parser.current_token().data == "=":
                # array definition and initialization - INT ID (LBRACK NUM RBRACK)+ ASSIGN ((LBRACE ... RBRACE)+|STRING) SEMICOLON
                self.parser.advance_token()  # skip ASSIGN

                if self.parser.current_token().type not in [Token.LBRACE, Token.STRING]:
                    raise BFSyntaxError("Expected LBRACE or STRING at '%s'" % self.parser.current_token())

                literal_tokens_list = self.parser.compile_array_initialization_list()
                self.parser.check_current_token_is(Token.SEMICOLON)
                self.parser.advance_token()  # skip SEMICOLON

                array_dimensions = get_variable_dimensions(variable)
                unpacked_literals_list = unpack_literal_tokens_to_array_dimensions(
                    ID_token, array_dimensions, literal_tokens_list
                )

                for literal in unpacked_literals_list:
                    code += get_literal_token_code(literal)  # evaluate this literal and point to next array element
                return code
            else:
                raise BFSyntaxError(
                    "Unexpected %s in array definition. Expected SEMICOLON (;) or ASSIGN (=)"
                    % self.parser.current_token()
                )

        elif self.parser.current_token().type == Token.SEMICOLON:  # no need to initialize
            self.parser.advance_token()  # skip SEMICOLON
            code += ">"  # advance to after this variable
        else:
            self.parser.check_current_token_is(Token.ASSIGN)
            if self.parser.current_token().data != "=":
                raise BFSyntaxError(
                    "Unexpected %s when initializing global variable. Expected ASSIGN (=)" % self.parser.current_token()
                )
            self.parser.advance_token()  # skip ASSIGN

            if not is_token_literal(self.parser.current_token()):
                raise BFSemanticError(
                    "Unexpected '%s'. expected literal (NUM | CHAR | TRUE | FALSE )" % str(self.parser.current_token())
                )

            code += get_literal_token_code(self.parser.current_token())

            self.parser.check_next_token_is(Token.SEMICOLON)
            self.parser.advance_token(amount=2)  # skip (NUM|CHAR|TRUE|FALSE) SEMICOLON

        return code

    def process_global_definitions(self):
        """
        Iterate through all tokens
        When encountering function definition - create Function object and pass it the function's tokens
        When encountering global variable definition - create Variable object
        Returns code that initializes global variables and advances the pointer to after them
        """
        code = ""
        token = self.parser.current_token()
        while token is not None and token.type in [Token.VOID, Token.INT, Token.SEMICOLON]:
            if token.type == Token.SEMICOLON:  # can have random semicolons ;)
                self.parser.advance_token()
                token = self.parser.current_token()
                continue
            self.parser.check_next_token_is(Token.ID)

            if self.parser.next_token(next_amount=2).type == Token.LPAREN:
                function = self.create_function_object()
                insert_function_object(function)
            elif token.type is Token.INT and self.parser.next_token(next_amount=2).type in [
                Token.SEMICOLON,
                Token.ASSIGN,
                Token.LBRACK,
            ]:
                code += self.compile_global_variable_definition()
            else:
                raise BFSyntaxError(
                    "Unexpected '%s' after '%s'. Expected '(' (function definition) or one of: '=', ';', '[' (global variable definition)"
                    % (str(self.parser.next_token(next_amount=2)), str(self.parser.next_token()))
                )

            token = self.parser.current_token()

        if self.parser.current_token() is not None:  # we have not reached the last token
            untouched_tokens = [str(t) for t in self.parser.tokens[self.parser.current_token_index :]]
            raise BFSyntaxError("Did not reach the end of the code. Untouched tokens:\n%s" % untouched_tokens)

        return code

    def compile(self):
        insert_library_functions()
        code = (
            self.process_global_definitions()
        )  # code that initializes global variables and advances pointer to after them

        check_function_exists(Token(Token.ID, 0, 0, "main"), 0)
        code += get_function_object("main").get_code(get_global_variables_size())
        code += "<" * get_global_variables_size()  # point to the first cell to end the program nicely :)
        return code


def compile(code, optimize_code=False):
    """
    :param code:  C-like code (string)
    :param optimize_code:  syntax optimization (bool)
    :return code:  Brainfuck code (string)
    """
    compiler = Compiler(code, optimize_code)
    brainfuck_code = compiler.compile()
    return brainfuck_code


if __name__ == "__main__":
    print("This file cannot be directly run")
    print("Please import it and use the 'compile' function")
    print("Which receives a C-like code (string) and returns Brainfuck code (string)")



"""Find the shortest path between a start and end point in a grid"""

from collections import deque
from dataclasses import dataclass
from random import Random
from typing import Any, Optional

from ..coaching import BaseCurriculum, RangeAttributeDefinition
from ..factory import ProceduralDataset, register_dataset

QUESTION_TEMPLATE = """Your task is to find the shortest path from the start to the destination point in a grid.

The grid is represented as a matrix with the following types of cells:
- *: your starting point
- #: your destination point
- O: an open cell
- X: a blocked cell

Therefore, you need to find the shortest path from * to #, moving only through open cells.

You may only move in four directions: up, down, left, and right.

If there is no path from * to #, simply write "infeasible" (without quotes).

Your output should be a sequence of directions that leads from * to #, e.g. right right down down up left

Now, find the shortest path from * to # in the following grid:
{grid}
"""

DATASET_NAME = "shortest_path"


@dataclass
class ShortestPathConfig:
    """Configuration for Shortest Path dataset generation"""

    min_rows: int = 5
    max_rows: int = 8
    min_cols: int = 5
    max_cols: int = 8
    p_blocked: float = 0.4

    size: int = 500  # Virtual dataset size
    seed: Optional[int] = None

    def validate(self):
        """Validate configuration parameters"""
        assert 1 <= self.min_rows, "min_rows must be at least 1"
        assert self.min_rows <= self.max_rows, "min_rows must be less than or equal to max_rows"
        assert 1 <= self.min_cols, "min_cols must be at least 1"
        assert self.min_cols <= self.max_cols, "min_cols must be less than or equal to max_cols"
        assert 0 <= self.p_blocked <= 1, "p_blocked must be between 0 and 1"


class ShortestPathDataset(ProceduralDataset):
    """Generates Shortest Path exercises with configurable difficulty"""

    def __init__(self, config: ShortestPathConfig):
        super().__init__(config=config, seed=config.seed, size=config.size)

    def _get_grid(self, rng: Random, rows: int, cols: int) -> list[list[str]]:
        """Generate a random grid with open and blocked cells"""
        grid = [["X" if rng.random() < self.config.p_blocked else "O" for _ in range(cols)] for _ in range(rows)]

        start_r, start_c = rng.randint(0, rows - 1), rng.randint(0, cols - 1)
        grid[start_r][start_c] = "*"

        while True:
            end_r, end_c = rng.randint(0, rows - 1), rng.randint(0, cols - 1)
            if (end_r, end_c) != (start_r, start_c):
                grid[end_r][end_c] = "#"
                break

        return grid

    def _matrix_to_str(self, matrix: list[list[int]]) -> str:
        """Get a string representation of the matrix"""
        return "\n".join(" ".join(str(x) for x in row) for row in matrix)

    def _get_answer(self, matrix: list[list[str]]) -> list[str]:
        """Run BFS to find the shortest path"""
        ROWS, COLS = len(matrix), len(matrix[0])
        DIRS = [(0, 1, "right"), (1, 0, "down"), (0, -1, "left"), (-1, 0, "up")]

        start_r, start_c = next((r, c) for r in range(ROWS) for c in range(COLS) if matrix[r][c] == "*")
        queue = deque([(start_r, start_c, [])])
        visited = set((start_r, start_c))

        while queue:
            r, c, path = queue.popleft()
            for dr, dc, direction in DIRS:
                new_r, new_c = r + dr, c + dc
                if 0 <= new_r < ROWS and 0 <= new_c < COLS and (new_r, new_c) not in visited:
                    new_path = path + [direction]
                    if matrix[new_r][new_c] == "#":
                        return new_path
                    if matrix[new_r][new_c] == "O":
                        visited.add((new_r, new_c))
                        queue.append((new_r, new_c, new_path))

        return []

    def _is_valid_path(self, matrix: list[list[str]], path: list[str]) -> bool:
        """Verifies the path goes from * to # without crossing X cells"""
        ROWS, COLS = len(matrix), len(matrix[0])
        DIRS = {"right": (0, 1), "down": (1, 0), "left": (0, -1), "up": (-1, 0)}

        start_r, start_c = next((r, c) for r in range(ROWS) for c in range(COLS) if matrix[r][c] == "*")
        end_r, end_c = next((r, c) for r in range(ROWS) for c in range(COLS) if matrix[r][c] == "#")

        r, c = start_r, start_c
        for direction in path:
            if direction not in DIRS:
                return False  # Invalid direction
            dr, dc = DIRS[direction]
            r, c = r + dr, c + dc
            if not (0 <= r < ROWS and 0 <= c < COLS):
                return False
            if matrix[r][c] == "X":
                return False

        return (r, c) == (end_r, end_c)

    def score_answer(self, answer: Optional[str], entry: dict[str, Any]) -> float:
        """Overwrite this method in derived classes if a single oracle answer is not available."""
        if isinstance(answer, str) and len(answer) > 0:
            oracle_answer = entry["answer"].strip()
            answer = answer.strip()

            # Exact answer
            if answer == oracle_answer:
                return 1.0

            matrix = entry["metadata"]["matrix"]
            answer = answer.split()
            oracle_answer = oracle_answer.split()

            # Path is valid and has the same length as the oracle answer
            if self._is_valid_path(matrix, answer) and len(answer) == len(oracle_answer):
                return 1.0

            # Path is valid but has a larger length than the oracle answer
            elif self._is_valid_path(matrix, answer):
                return 0.5

        return 0.0

    def __getitem__(self, idx: int) -> dict:
        """Generate a single Shortest Path question"""
        rng = Random(self.seed + idx)

        rows = rng.randint(self.config.min_rows, self.config.max_rows)
        cols = rng.randint(self.config.min_cols, self.config.max_cols)
        matrix = self._get_grid(rng, rows, cols)
        matrix_str = self._matrix_to_str(matrix)
        answer = self._get_answer(matrix)
        answer_str = " ".join(answer) if answer else "infeasible"

        return {
            "question": QUESTION_TEMPLATE.format(grid=matrix_str),
            "answer": answer_str,
            "metadata": {
                "source_dataset": DATASET_NAME,
                "source_index": idx,
                "matrix": matrix,
                "solution": answer,
                "difficulty": {
                    "rows": (self.config.min_rows, self.config.max_rows),
                    "cols": (self.config.min_cols, self.config.max_cols),
                },
            },
        }


class ShortestPathCurriculum(BaseCurriculum):
    def __init__(self):
        super().__init__(ShortestPathCurriculum.__name__, ShortestPathConfig)

        # Define attributes
        self._define_attributes(
            RangeAttributeDefinition(
                name="rows",
                levels=[10, 25, 50, 100],
                description="Number of rows in the grid",
                lower_field_name="min_rows",
                upper_field_name="max_rows",
            ),
            RangeAttributeDefinition(
                name="cols",
                levels=[10, 25, 50, 100],
                description="Number of columns in the grid",
                lower_field_name="min_cols",
                upper_field_name="max_cols",
            ),
        )


register_dataset(DATASET_NAME, ShortestPathDataset, ShortestPathConfig, ShortestPathCurriculum)



"""Leg counting task generator"""

from dataclasses import dataclass
from random import Random
from typing import Optional

from ..coaching import BaseCurriculum, RangeAttributeDefinition
from ..factory import ProceduralDataset, register_dataset

ANIMALS = {
    # Animals with 0 legs
    "snake": 0,
    "sea slug": 0,
    "jellyfish": 0,
    "flatworm": 0,
    "leech": 0,
    # Animals with 2 legs
    "chicken": 2,
    "bird": 2,
    "human": 2,
    "duck": 2,
    # Animals with 4 legs
    "dog": 4,
    "cat": 4,
    "cow": 4,
    "horse": 4,
    "lion": 4,
    "elephant": 4,
    "giraffe": 4,
    "tiger": 4,
    "deer": 4,
    "sheep": 4,
    # Animals with 5 legs
    "starfish": 5,
    # Animals with 6 legs
    "insect": 6,
    "ant": 6,
    "butterfly": 6,
    "beetle": 6,
    "bee": 6,
    "wasp": 6,
    "grasshopper": 6,
    "cricket": 6,
    "cockroach": 6,
    "praying mantis": 6,
    "firefly": 6,
    # Animals with 8 legs
    "spider": 8,
    "scorpion": 8,
    # Animals with 10 legs
    "crab": 10,
    "lobster": 10,
    "shrimp": 10,
    # Animals with 14 legs
    "woodlouse": 14,
}

QUESTION_TEMPLATE = """Your task is to count how many legs there are in total when given a list of animals.

Now, how many legs are there in total if you have {animals}?
"""

DATASET_NAME = "leg_counting"


@dataclass
class LegCountingConfig:
    """Configuration for leg counting task generation"""

    min_animals: int = 3  # Minimum number of animals in problem
    max_animals: int = 10  # Maximum number of animals
    min_instances: int = 1  # Minimum instances of each animal
    max_instances: int = 15  # Maximum instances of each animal
    seed: Optional[int] = None
    size: int = 500  # Virtual dataset size

    def validate(self) -> None:
        """Validate configuration parameters"""
        assert self.min_animals > 0, "min_animals must be positive"
        assert self.max_animals >= self.min_animals, "max_animals must be >= min_animals"
        assert self.max_animals <= len(ANIMALS), "max_animals must be <= number of available animals"  # 37
        assert self.min_instances > 0, "min_instances must be positive"
        assert self.max_instances >= self.min_instances, "max_instances must be >= min_instances"


class LegCountingDataset(ProceduralDataset):
    """Generates leg counting arithmetic tasks"""

    def __init__(self, config: LegCountingConfig):
        super().__init__(config=config, seed=config.seed, size=config.size)

    def _generate_animals(self, rng: Random) -> dict[str, int]:
        """Generate a random set of animals and their counts"""
        num_types = rng.randint(self.config.min_animals, self.config.max_animals)
        animals = {}

        # Select random animals
        selected_animals = rng.sample(list(ANIMALS.keys()), num_types)
        for animal in selected_animals:
            count = rng.randint(self.config.min_instances, self.config.max_instances)
            animals[animal] = count

        return animals

    def __getitem__(self, idx: int) -> dict:
        """Generate a single leg counting task"""
        rng = Random(self.seed + idx)

        # Generate random animals and their counts
        animals = self._generate_animals(rng)

        # Calculate total legs
        total_legs = sum(count * ANIMALS[animal] for animal, count in animals.items())

        # Format animal counts for question
        animal_list = []
        for animal, count in animals.items():
            animal_list.append(f"{count} {animal}{'s' if count > 1 else ''}")

        return {
            "question": QUESTION_TEMPLATE.format(animals=", ".join(animal_list)),
            "answer": str(total_legs),
            "metadata": {
                "source_dataset": DATASET_NAME,
                "source_index": idx,
                "animals": animals,
                "num_animals": len(animals),
                "total_legs": total_legs,
                "difficulty": {
                    "num_animals": (self.config.min_animals, self.config.max_animals),
                    "num_instances": (self.config.min_instances, self.config.max_instances),
                },
            },
        }


class LegCountingCurriculum(BaseCurriculum):
    def __init__(self):
        super().__init__(LegCountingCurriculum.__name__, LegCountingConfig)

        # Define attributes
        self._define_attributes(
            RangeAttributeDefinition(
                name="num_animals",
                levels=list(range(1, 37)),
                description="Number of animals in question",
                lower_field_name="min_animals",
                upper_field_name="max_animals",
            ),
            RangeAttributeDefinition(
                name="num_instances",
                levels=[2, 4, 8, 16, 32, 64, 128, 256, 512, 1024],
                description="Number of instances of each animal",
                lower_field_name="min_instances",
                upper_field_name="max_instances",
                ensure_interval=True,
            ),
        )


register_dataset(DATASET_NAME, LegCountingDataset, LegCountingConfig, LegCountingCurriculum)



"""Registry for managing active experiments."""

from typing import Optional

from ..composite import CompositeConfig
from .experiment import Experiment


class ExperimentRegistry:
    """Singleton registry for managing active experiments."""

    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance._experiments = {}
        return cls._instance

    def register_experiment(self, name: str, config: CompositeConfig) -> None:
        """Register a new experiment with the given name and configuration."""
        self._experiments[name] = Experiment.create(name, config)

    def get_experiment(self, name: str) -> Optional[Experiment]:
        """Get an experiment by name."""
        return self._experiments.get(name)

    def list_experiments(self) -> list[str]:
        """List all registered experiment names."""
        return list(self._experiments.keys())

    def remove_experiment(self, name: str) -> bool:
        """Remove an experiment by name. Returns True if removed, False if not found."""
        return bool(self._experiments.pop(name, None))



"""Given a string, return all possible partitions of the string such that each substring is a palindrome.

A popular Leetcode problem:
https://leetcode.com/problems/palindrome-partitioning/description/
"""

import json
import string
from dataclasses import dataclass
from random import Random
from typing import Any, Optional

from ..coaching import BaseCurriculum, RangeAttributeDefinition
from ..factory import ProceduralDataset, register_dataset

QUESTION_TEMPLATE = """Given a string, partition it such that every substring is a palindrome.

A palindrome is a word that reads the same backward as forward.

You may return all possible palindrome partitioning in any order.

Your output should be a list of lists, where each list represents a palindrome partition, e.g. [["a","a","b"],["aa","b"]].

Partition the following string into palindromes: {string}
"""

DATASET_NAME = "palindrome_partitioning"


@dataclass
class PalindromePartitioningConfig:
    """Configuration for Palindrome Partitioning dataset generation"""

    min_string_len: int = 5
    max_string_len: int = 15
    min_substring_palindrome_len: int = 1
    max_substring_palindrome_len: int = 5

    size: int = 500  # Virtual dataset size
    seed: Optional[int] = None

    def validate(self):
        """Validate configuration parameters"""
        assert 1 <= self.min_string_len, "Minimum string length must be at least 1"
        assert self.min_string_len <= self.max_string_len, "Minimum string length must be less than or equal to maximum"
        assert 1 <= self.min_substring_palindrome_len, "Minimum substring palindrome length must be at least 1"
        assert (
            self.min_substring_palindrome_len <= self.max_substring_palindrome_len
        ), "Minimum substring palindrome length must be less than or equal to maximum"
        assert (
            self.max_substring_palindrome_len <= self.max_string_len
        ), "Maximum substring palindrome length must be less than or equal to maximum string length"


class PalindromePartitioningDataset(ProceduralDataset):
    """Generates Palindrome Partitioning exercises with configurable difficulty"""

    def __init__(self, config: PalindromePartitioningConfig):
        super().__init__(config=config, seed=config.seed, size=config.size)

    def _sort_list(self, lst: list[list[str]]) -> list[list[str]]:
        """Sort the list of palindrome partitions"""
        return sorted(lst, key=lambda x: x[0] if x else "")

    def to_set_of_tuples(self, list_of_lists: list[list[str]]) -> set[tuple[str]]:
        """Convert a list of lists to a set of tuples"""
        return {tuple(lst) for lst in list_of_lists}

    def _palindrome_partitioning(self, string: str) -> list[list[str]]:
        """Return all possible palindrome partitions of a string"""
        if not string:
            return []
        dp = {}

        def is_palindrome(i, j) -> bool:
            if i >= j:
                return True
            if (i, j) in dp:
                return dp[(i, j)]
            dp[(i, j)] = string[i] == string[j] and is_palindrome(i + 1, j - 1)
            return dp[(i, j)]

        res, temp = [], []

        def _partition(idx) -> None:
            if idx >= len(string):
                res.append(temp[:])
            for i in range(idx, len(string)):
                if is_palindrome(idx, i):
                    temp.append(string[idx : i + 1])
                    _partition(i + 1)
                    temp.pop()

        _partition(0)
        return self._sort_list(res)

    def score_answer(self, answer: Optional[str], entry: dict[str, Any]) -> float:
        """Score a single Palindrome Partitioning question"""
        if answer is not None:
            try:
                answer = self.to_set_of_tuples(json.loads(answer))
                oracle = self.to_set_of_tuples(entry["metadata"]["solution"])
                if answer == oracle:
                    return 1.0
            except Exception:
                pass
        return 0.0

    def _generate_palindrome_letters(self, rng: Random, length: int) -> list[str]:
        """Generate a set of letters that can form a palindrome."""
        half_length = length // 2
        letters = rng.choices(string.ascii_lowercase, k=half_length)
        if length % 2 == 1:
            middle_letter = rng.choice(string.ascii_lowercase)
            return letters + [middle_letter] + letters[::-1]
        return letters + letters[::-1]

    def _get_string(self, rng: Random, string_len: int) -> str:
        """Generate a random string"""
        output = ""
        while len(output) < string_len:
            palindrome_len = min(
                string_len - len(output),
                rng.randint(self.config.min_substring_palindrome_len, self.config.max_substring_palindrome_len),
            )
            substring = "".join(self._generate_palindrome_letters(rng, palindrome_len))
            output += substring
        return output

    def __getitem__(self, idx: int) -> dict:
        """Generate a single Palindrome Partitioning question"""
        rng = Random(self.seed + idx)

        string_len = rng.randint(self.config.min_string_len, self.config.max_string_len)
        string = self._get_string(rng, string_len)
        answer = self._palindrome_partitioning(string)
        answer_str = json.dumps(answer)

        return {
            "question": QUESTION_TEMPLATE.format(string=string),
            "answer": answer_str,
            "metadata": {
                "source_dataset": DATASET_NAME,
                "source_index": idx,
                "string": string,
                "solution": answer,
                "string_len": string_len,
                "difficulty": {
                    "string_len": (self.config.min_string_len, self.config.max_string_len),
                    "substring_palindrome_len": (
                        self.config.min_substring_palindrome_len,
                        self.config.max_substring_palindrome_len,
                    ),
                },
            },
        }


class PalindromePartitioningCurriculum(BaseCurriculum):
    def __init__(self):
        super().__init__(PalindromePartitioningCurriculum.__name__, PalindromePartitioningConfig)

        # Define attributes
        self._define_attributes(
            RangeAttributeDefinition(
                name="string_len",
                levels=[1, 5, 10, 15],
                description="Length of the string",
                lower_field_name="min_string_len",
                upper_field_name="max_string_len",
                ensure_interval=True,
            ),
            RangeAttributeDefinition(
                name="substring_palindrome_len",
                levels=[1, 3, 5, 7],
                description="Length of the substring palindrome",
                lower_field_name="min_substring_palindrome_len",
                upper_field_name="max_substring_palindrome_len",
                ensure_interval=True,
            ),
        )


register_dataset(
    DATASET_NAME,
    PalindromePartitioningDataset,
    PalindromePartitioningConfig,
    PalindromePartitioningCurriculum,
)
